#include "i2cmaster.h"
#include "irmp/irmp.h"
#include "hw.h"
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>

#define _BV(bit) (1 << (bit)) // _BV macro is not defined so define it here
#define BTNF() (bit_is_clear(PINC, 3))
#define BTNR() (bit_is_clear(PINC, 2))
#define BTNFF() (bit_is_clear(PINC, 1))
#define BTNRR() (bit_is_clear(PINC, 0))
#define BTNPROGRAM() (bit_is_clear(PINB, 5))
#define TUNING_STEP 1 // 1 = 50 kHz
#define TUNING_STEP_FAST 10 // 10 = 500 kHz
#define BASE_FREQ 2495 // 87.50 MHz
#define BASE_FREQ_MHZ 8750 // 87.50 MHz
#define MAX_FREQ 2905 // 108.00 MHz
#define IR_LEARNING_PROGRAM0 1
#define IR_LEARNING_PROGRAM1 2
#define IR_LEARNING_PROGRAM2 3
#define IR_LEARNING_PROGRAM3 4
#define IR_LEARNING_PROGRAM4 5
#define IR_LEARNING_PROGRAM5 6
#define IR_LEARNING_PROGRAM6 7
#define IR_LEARNING_PROGRAM7 8
#define IR_LEARNING_PROGRAM8 9
#define IR_LEARNING_PROGRAM9 10
#define IR_LEARNING_FF 11
#define IR_LEARNING_F 12
#define IR_LEARNING_RR 13
#define IR_LEARNING_R 14
#define IR_SET_PROGRAM0 1
#define IR_SET_PROGRAM1 2
#define IR_SET_PROGRAM2 3
#define IR_SET_PROGRAM3 4
#define IR_SET_PROGRAM4 5
#define IR_SET_PROGRAM5 6
#define IR_SET_PROGRAM6 7
#define IR_SET_PROGRAM7 8
#define IR_SET_PROGRAM8 9
#define IR_SET_PROGRAM9 10

volatile unsigned int freq;
volatile unsigned int freqMhz;
volatile unsigned char activeseg = 0; // indicates active display segment (used for multiplexing)
volatile unsigned int lastFreqSaveCheck = 0;
volatile unsigned int lastFreq = 0;
volatile unsigned char irLearningMode = 0;
volatile unsigned char isIrProgramSetMode = 0;
const unsigned char* LAST_FREQ_ADDRESS = (unsigned char*)0;
static IRMP_DATA irmp_data;
static IRMP_DATA irmp_program0;
static unsigned int freqProgram0;
const unsigned char* EEPROM_ADDRESS_PROGRAM0 = (unsigned char*)100;
const unsigned char* EEPROM_ADDRESS_PROGRAM0_FREQ = (unsigned char*)106; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program1;
static unsigned int freqProgram1;
const unsigned char* EEPROM_ADDRESS_PROGRAM1 = (unsigned char*)110;
const unsigned char* EEPROM_ADDRESS_PROGRAM1_FREQ = (unsigned char*)116; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program2;
static unsigned int freqProgram2;
const unsigned char* EEPROM_ADDRESS_PROGRAM2 = (unsigned char*)120;
const unsigned char* EEPROM_ADDRESS_PROGRAM2_FREQ = (unsigned char*)126; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program3;
static unsigned int freqProgram3;
const unsigned char* EEPROM_ADDRESS_PROGRAM3 = (unsigned char*)130;
const unsigned char* EEPROM_ADDRESS_PROGRAM3_FREQ = (unsigned char*)136; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program4;
static unsigned int freqProgram4;
const unsigned char* EEPROM_ADDRESS_PROGRAM4 = (unsigned char*)140;
const unsigned char* EEPROM_ADDRESS_PROGRAM4_FREQ = (unsigned char*)146; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program5;
static unsigned int freqProgram5;
const unsigned char* EEPROM_ADDRESS_PROGRAM5 = (unsigned char*)150;
const unsigned char* EEPROM_ADDRESS_PROGRAM5_FREQ = (unsigned char*)156; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program6;
static unsigned int freqProgram6;
const unsigned char* EEPROM_ADDRESS_PROGRAM6 = (unsigned char*)160;
const unsigned char* EEPROM_ADDRESS_PROGRAM6_FREQ = (unsigned char*)166; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program7;
static unsigned int freqProgram7;
const unsigned char* EEPROM_ADDRESS_PROGRAM7 = (unsigned char*)170;
const unsigned char* EEPROM_ADDRESS_PROGRAM7_FREQ = (unsigned char*)176; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program8;
static unsigned int freqProgram8;
const unsigned char* EEPROM_ADDRESS_PROGRAM8 = (unsigned char*)180;
const unsigned char* EEPROM_ADDRESS_PROGRAM8_FREQ = (unsigned char*)186; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program9;
static unsigned int freqProgram9;
const unsigned char* EEPROM_ADDRESS_PROGRAM9 = (unsigned char*)190;
const unsigned char* EEPROM_ADDRESS_PROGRAM9_FREQ = (unsigned char*)196; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_rr;
const unsigned char* EEPROM_ADDRESS_RR = (unsigned char*)200;
static IRMP_DATA irmp_r;
const unsigned char* EEPROM_ADDRESS_R = (unsigned char*)206;
static IRMP_DATA irmp_ff;
const unsigned char* EEPROM_ADDRESS_FF = (unsigned char*)212;
static IRMP_DATA irmp_f;
const unsigned char* EEPROM_ADDRESS_F = (unsigned char*)218;

// display segments configuration
#define _s_A 0
#define _s_B 1
#define _s_C 2
#define _s_D 3
#define _s_E 4
#define _s_F 5
#define _s_G 6
#define _s_dot 7

const unsigned char segs[] =
{
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F), //0
	_BV(_s_B) | _BV(_s_C), //1
	_BV(_s_A) | _BV(_s_B) | _BV(_s_D) | _BV(_s_E) | _BV(_s_G), //2
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_G), //3
	_BV(_s_B) | _BV(_s_C) | _BV(_s_F) | _BV(_s_G), //4
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G), //5
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //6
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C), //7
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G),//8
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G),//9
	0, // empty 10
	_BV(_s_A) | _BV(_s_B) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G),//P (11)
	_BV(_s_A) | _BV(_s_B) | _BV(_s_E) | _BV(_s_G),//? (12)
	_BV(_s_A) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G),//F (13)
	_BV(_s_E) | _BV(_s_G),//R (14)
};

static void setFreq()
{
	unsigned char dividerH = freq >> 8;
	unsigned char dividerL = freq;
	
	freqMhz = BASE_FREQ_MHZ + ((freq - BASE_FREQ) * 5); // 1 in 'freq' == 50 kHz
	
	i2c_start_wait(0xc2 + I2C_WRITE); // set device address and write mode
	i2c_write(dividerH); // program divider byte 1 (msb)
	i2c_write(dividerL); // program divider byte 2 (lsb)
	i2c_write(0b11001000); // control byte
	i2c_write(0b00000001); // switch to low band
	i2c_stop();
	
	lastFreqSaveCheck = 0; // invoke last frequency save to eeprom after a few seconds to avoid eeprom wear
}

static void fastBack_rr()
{
	if (freq == BASE_FREQ)
	{
		freq = MAX_FREQ;
	} else if (freq - TUNING_STEP_FAST < BASE_FREQ) {
		freq = BASE_FREQ;
	} else {
		freq -= TUNING_STEP_FAST;
	}

	setFreq();
}

static void back_r()
{
	if (freq == BASE_FREQ)
	{
		freq = MAX_FREQ;
	} else if (freq - TUNING_STEP < BASE_FREQ) {
		freq = BASE_FREQ;
	} else {
		freq -= TUNING_STEP;
	}

	setFreq();
}

static void forward_f()
{
	if (freq == MAX_FREQ)
	{
		freq = BASE_FREQ;
	} else if (freq + TUNING_STEP > MAX_FREQ) {
		freq = MAX_FREQ;
	} else {
		freq += TUNING_STEP;
	}

	setFreq();
}

static void fastForward_ff()
{
	if (freq == MAX_FREQ)
	{
		freq = BASE_FREQ;
	} else if (freq + TUNING_STEP_FAST > MAX_FREQ) {
		freq = MAX_FREQ;
	} else {
		freq += TUNING_STEP_FAST;
	}

	setFreq();
}

static void irmp_timer_init()
{
	#if defined (__AVR_ATtiny45__) || defined (__AVR_ATtiny85__)                // ATtiny45 / ATtiny85:

	#if F_CPU >= 16000000L
	OCR1C   =  (F_CPU / F_INTERRUPTS / 8) - 1;                              // compare value: 1/15000 of CPU frequency, presc = 8
	TCCR1   = (1 << CTC1) | (1 << CS12);                                    // switch CTC Mode on, set prescaler to 8
	#else
	OCR1C   =  (F_CPU / F_INTERRUPTS / 4) - 1;                              // compare value: 1/15000 of CPU frequency, presc = 4
	TCCR1   = (1 << CTC1) | (1 << CS11) | (1 << CS10);                      // switch CTC Mode on, set prescaler to 4
	#endif

	#else                                                                       // ATmegaXX:
	OCR1A   =  (F_CPU / F_INTERRUPTS) - 1;                                  // compare value: 1/15000 of CPU frequency
	TCCR1B  = (1 << WGM12) | (1 << CS10);                                   // switch CTC Mode on, set prescaler to 1
	#endif

	#ifdef TIMSK1
	TIMSK1  = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#else
	TIMSK   = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#endif
}

#ifdef TIM1_COMPA_vect                                                      // ATtiny84
#define COMPA_VECT  TIM1_COMPA_vect
#else
#define COMPA_VECT  TIMER1_COMPA_vect                                       // ATmega
#endif

static void learnIrCode(int learningMode, const unsigned char* eepromAddress, IRMP_DATA* irmpDataVariable)
{
	IRMP_DATA irmp_learning_data;
	memset(&irmp_learning_data, 0x00, sizeof(irmp_learning_data));
	irLearningMode = learningMode;
	_delay_ms(1000);
	while (!irmp_get_data(&irmp_learning_data) || (irmp_learning_data.flags & IRMP_FLAG_REPETITION));
	eeprom_write_block(&irmp_learning_data, (unsigned char *)eepromAddress, sizeof(irmp_learning_data));
	memcpy(irmpDataVariable, &irmp_learning_data, sizeof(irmp_learning_data));
}

int main(void)
{
	// PORTD: segments
	// PORTB: common anodes (PB0-PB3)
	
	// set all PORTD pins as output, PB0-PB3 as well as outputs, PB4 as input, PB5 as input
	DDRD = 0xFF;
	DDRB |= (1 << DDB0) | (1 << DDB1) | (1 << DDB2) | (1 << DDB3);
	DDRB &=~ (1 << PB4);
	DDRB &=~ (1 << PB5);
	
	// set PC0 and PC1 as input
	DDRC &=~ (1 << PC0);
	DDRC &=~ (1 << PC1);
	PORTC |= (1 << PC1) | (1 << PC0);
  
	// show empty display at the beginning
	PORTB = 0xFF;
	PORTD = 0xFF;
	
	// TIMER2 configuration
	TIMSK2 |= _BV(TOIE2); // enable overflow interrupt
	TCCR2B |= _BV(CS21); // configure prescaler to 8 and start timer

	_delay_ms(200);
	i2c_init();
	memset(&irmp_data, 0x00, sizeof(irmp_data));
	irmp_init();
	irmp_timer_init();
	sei(); // enable global interrupts
	
	unsigned int r;
	eeprom_read_block(&r, (unsigned char *)LAST_FREQ_ADDRESS, sizeof(r));
	lastFreq = r;
    freq = (lastFreq >= BASE_FREQ && lastFreq <= MAX_FREQ) ? lastFreq : BASE_FREQ;
	
	setFreq();
	
	// read trained ir codes and frequencies
	eeprom_read_block(&irmp_program0, (unsigned char *)EEPROM_ADDRESS_PROGRAM0, sizeof(irmp_program0));
	eeprom_read_block(&freqProgram0, (unsigned char *)EEPROM_ADDRESS_PROGRAM0_FREQ, sizeof(freqProgram0));
	eeprom_read_block(&irmp_program1, (unsigned char *)EEPROM_ADDRESS_PROGRAM1, sizeof(irmp_program1));
	eeprom_read_block(&freqProgram1, (unsigned char *)EEPROM_ADDRESS_PROGRAM1_FREQ, sizeof(freqProgram1));
	eeprom_read_block(&irmp_program2, (unsigned char *)EEPROM_ADDRESS_PROGRAM2, sizeof(irmp_program2));
	eeprom_read_block(&freqProgram2, (unsigned char *)EEPROM_ADDRESS_PROGRAM2_FREQ, sizeof(freqProgram2));
	eeprom_read_block(&irmp_program3, (unsigned char *)EEPROM_ADDRESS_PROGRAM3, sizeof(irmp_program3));
	eeprom_read_block(&freqProgram3, (unsigned char *)EEPROM_ADDRESS_PROGRAM3_FREQ, sizeof(freqProgram3));
	eeprom_read_block(&irmp_program4, (unsigned char *)EEPROM_ADDRESS_PROGRAM4, sizeof(irmp_program4));
	eeprom_read_block(&freqProgram4, (unsigned char *)EEPROM_ADDRESS_PROGRAM4_FREQ, sizeof(freqProgram4));
	eeprom_read_block(&irmp_program5, (unsigned char *)EEPROM_ADDRESS_PROGRAM5, sizeof(irmp_program5));
	eeprom_read_block(&freqProgram5, (unsigned char *)EEPROM_ADDRESS_PROGRAM5_FREQ, sizeof(freqProgram5));
	eeprom_read_block(&irmp_program6, (unsigned char *)EEPROM_ADDRESS_PROGRAM6, sizeof(irmp_program6));
	eeprom_read_block(&freqProgram6, (unsigned char *)EEPROM_ADDRESS_PROGRAM6_FREQ, sizeof(freqProgram6));
	eeprom_read_block(&irmp_program7, (unsigned char *)EEPROM_ADDRESS_PROGRAM7, sizeof(irmp_program7));
	eeprom_read_block(&freqProgram7, (unsigned char *)EEPROM_ADDRESS_PROGRAM7_FREQ, sizeof(freqProgram7));
	eeprom_read_block(&irmp_program8, (unsigned char *)EEPROM_ADDRESS_PROGRAM8, sizeof(irmp_program8));
	eeprom_read_block(&freqProgram8, (unsigned char *)EEPROM_ADDRESS_PROGRAM8_FREQ, sizeof(freqProgram8));
	eeprom_read_block(&irmp_program9, (unsigned char *)EEPROM_ADDRESS_PROGRAM9, sizeof(irmp_program9));
	eeprom_read_block(&freqProgram9, (unsigned char *)EEPROM_ADDRESS_PROGRAM9_FREQ, sizeof(freqProgram9));
	eeprom_read_block(&irmp_rr, (unsigned char *)EEPROM_ADDRESS_RR, sizeof(irmp_rr));
	eeprom_read_block(&irmp_r, (unsigned char *)EEPROM_ADDRESS_R, sizeof(irmp_r));
	eeprom_read_block(&irmp_ff, (unsigned char *)EEPROM_ADDRESS_FF, sizeof(irmp_ff));
	eeprom_read_block(&irmp_f, (unsigned char *)EEPROM_ADDRESS_F, sizeof(irmp_f));

    for (;;)
    {
		if (BTNF() && BTNR())
		{
			// enter IR code learning mode
			learnIrCode(IR_LEARNING_PROGRAM0, EEPROM_ADDRESS_PROGRAM0, &irmp_program0);
			learnIrCode(IR_LEARNING_PROGRAM1, EEPROM_ADDRESS_PROGRAM1, &irmp_program1);
			learnIrCode(IR_LEARNING_PROGRAM2, EEPROM_ADDRESS_PROGRAM2, &irmp_program2);
			learnIrCode(IR_LEARNING_PROGRAM3, EEPROM_ADDRESS_PROGRAM3, &irmp_program3);
			learnIrCode(IR_LEARNING_PROGRAM4, EEPROM_ADDRESS_PROGRAM4, &irmp_program4);
			learnIrCode(IR_LEARNING_PROGRAM5, EEPROM_ADDRESS_PROGRAM5, &irmp_program5);
			learnIrCode(IR_LEARNING_PROGRAM6, EEPROM_ADDRESS_PROGRAM6, &irmp_program6);
			learnIrCode(IR_LEARNING_PROGRAM7, EEPROM_ADDRESS_PROGRAM7, &irmp_program7);
			learnIrCode(IR_LEARNING_PROGRAM8, EEPROM_ADDRESS_PROGRAM8, &irmp_program8);
			learnIrCode(IR_LEARNING_PROGRAM9, EEPROM_ADDRESS_PROGRAM9, &irmp_program9);
			learnIrCode(IR_LEARNING_RR, EEPROM_ADDRESS_RR, &irmp_rr);
			learnIrCode(IR_LEARNING_R, EEPROM_ADDRESS_R, &irmp_r);
			learnIrCode(IR_LEARNING_FF, EEPROM_ADDRESS_FF, &irmp_ff);
			learnIrCode(IR_LEARNING_F, EEPROM_ADDRESS_F, &irmp_f);
		
			irLearningMode = 0;
			_delay_ms(1000);
		}
		else if (BTNPROGRAM())
		{
			// save current frequency, but to which program? get IR code that should be already learned
			isIrProgramSetMode = 1;
			IRMP_DATA irmp_set_data;
			memset(&irmp_set_data, 0x00, sizeof(irmp_set_data));
			_delay_ms(500); // so that this mode is not canceled right away
			unsigned char wasRecognized = 0;
			unsigned char canceled = 0;
		
			do 
			{
				while (!irmp_get_data(&irmp_set_data))
				{
					if (BTNPROGRAM())
					{
						canceled = 1;
						break;
					}
				}

				if (canceled == 0)
				{
					if (memcmp(&irmp_set_data, &irmp_program0, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM0_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM0_FREQ + 1, freq >> 8);
						freqProgram0 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program1, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM1_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM1_FREQ + 1, freq >> 8);
						freqProgram1 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program2, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM2_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM2_FREQ + 1, freq >> 8);
						freqProgram2 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program3, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM3_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM3_FREQ + 1, freq >> 8);
						freqProgram3 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program4, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM4_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM4_FREQ + 1, freq >> 8);
						freqProgram4 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program5, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM5_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM5_FREQ + 1, freq >> 8);
						freqProgram5 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program6, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM6_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM6_FREQ + 1, freq >> 8);
						freqProgram6 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program7, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM7_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM7_FREQ + 1, freq >> 8);
						freqProgram7 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program8, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM8_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM8_FREQ + 1, freq >> 8);
						freqProgram8 = freq;
						wasRecognized = 1;
					}
					else if (memcmp(&irmp_set_data, &irmp_program9, sizeof(irmp_set_data) - 1) == 0)
					{
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM9_FREQ, freq & 0xFF);
						eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM9_FREQ + 1, freq >> 8);
						freqProgram9 = freq;
						wasRecognized = 1;
					}
				}
			} while (wasRecognized == 0 && canceled == 0);
			
			isIrProgramSetMode = 0;
			_delay_ms(500); // without delay, when canceling, this mode would be entered again
		}
		else if (BTNF())
		{
			forward_f();
			_delay_ms(200);
		}
		else if (BTNR())
		{
			back_r();
			_delay_ms(200);
		}
		else if (BTNFF())
		{
			fastForward_ff();
			_delay_ms(200);
		}
		else if (BTNRR())
		{
			fastBack_rr();
			_delay_ms(200);
		}
		
		if (irmp_get_data(&irmp_data) && !(irmp_data.flags & IRMP_FLAG_REPETITION)) // repetition is not supported because it's unwantedly triggered even on single button press
		{
			/* ir signal decoded, do something here... */
			/* irmp_data.protocol is the protocol, see irmp.h */
			/* irmp_data.address is the address/manufacturer code of ir sender */
			/* irmp_data.command is the command code */
			
			// compare trained code with received code
			// below we compare sizeof - 1 because we want to exclude 'flags' property (it indicates repetition)
			if ((freqProgram0 >= BASE_FREQ && freqProgram0 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program0, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram0;
				setFreq();
			}
			else if ((freqProgram1 >= BASE_FREQ && freqProgram1 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program1, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram1;
				setFreq();
			}
			else if ((freqProgram2 >= BASE_FREQ && freqProgram2 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program2, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram2;
				setFreq();
			}
			else if ((freqProgram3 >= BASE_FREQ && freqProgram3 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program3, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram3;
				setFreq();
			}
			else if ((freqProgram4 >= BASE_FREQ && freqProgram4 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program4, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram4;
				setFreq();
			}
			else if ((freqProgram5 >= BASE_FREQ && freqProgram5 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program5, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram5;
				setFreq();
			}
			else if ((freqProgram6 >= BASE_FREQ && freqProgram6 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program6, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram6;
				setFreq();
			}
			else if ((freqProgram7 >= BASE_FREQ && freqProgram7 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program7, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram7;
				setFreq();
			}
			else if ((freqProgram8 >= BASE_FREQ && freqProgram8 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program8, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram8;
				setFreq();
			}
			else if ((freqProgram9 >= BASE_FREQ && freqProgram9 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program9, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram9;
				setFreq();
			}
			else if (memcmp(&irmp_data, &irmp_ff, sizeof(irmp_data) - 1) == 0)
			{
				fastForward_ff();
			}
			else if (memcmp(&irmp_data, &irmp_f, sizeof(irmp_data) - 1) == 0)
			{
				forward_f();
			}
			else if (memcmp(&irmp_data, &irmp_rr, sizeof(irmp_data) - 1) == 0)
			{
				fastBack_rr();
			}
			else if (memcmp(&irmp_data, &irmp_r, sizeof(irmp_data) - 1) == 0)
			{
				back_r();
			}
		}
			
		_delay_ms(10);
	}
}

ISR(TIMER2_OVF_vect)
{
	if (activeseg == 3)
	{
		activeseg = 0;
	} else {
		activeseg++;
	}
	
	unsigned char seg1 = 10;
	unsigned char seg2 = 10;
	unsigned char seg3 = 10;
	unsigned char seg4 = 10;
	
	if (isIrProgramSetMode == 1) {
		seg1 = 11; // P
		seg2 = 12; // 0
		seg3 = 10; // empty
		seg4 = 10; // empty
    }
	else if (irLearningMode != 0)
	{
		switch (irLearningMode)
		{
			case IR_LEARNING_PROGRAM0:
				seg1 = 11; // P
				seg2 = 0; // 0
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM1:
				seg1 = 11; // P
				seg2 = 1; // 1
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM2:
				seg1 = 11; // P
				seg2 = 2; // 2
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM3:
				seg1 = 11; // P
				seg2 = 3; // 3
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM4:
				seg1 = 11; // P
				seg2 = 4; // 4
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM5:
				seg1 = 11; // P
				seg2 = 5; // 5
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM6:
				seg1 = 11; // P
				seg2 = 6; // 6
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM7:
				seg1 = 11; // P
				seg2 = 7; // 7
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM8:
				seg1 = 11; // P
				seg2 = 8; // 8
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_PROGRAM9:
				seg1 = 11; // P
				seg2 = 9; // 9
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_FF:
				seg1 = 13; // F
				seg2 = 13; // F
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_F:
				seg1 = 13; // F
				seg2 = 10; // empty
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_RR:
				seg1 = 14; // R
				seg2 = 14; // R
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
			case IR_LEARNING_R:
				seg1 = 14; // R
				seg2 = 10; // empty
				seg3 = 10; // empty
				seg4 = 12; // ?
			break;
		}
	}
	else if (irLearningMode == 0 && isIrProgramSetMode == 0)
	{
		if (freqMhz < 10000)
		{
			seg1 = freqMhz / 1000;
			seg2 = (freqMhz % 1000) / 100;
			seg3 = (freqMhz % 100) / 10;
			seg4 = freqMhz % 10;
	    } else {
			seg1 =  1;
			seg2 = (freqMhz % 10000) / 1000;
			seg3 = (freqMhz % 1000) / 100;
			seg4 = (freqMhz % 100) / 10;
		}
	}
	
	if (lastFreqSaveCheck == 60000) // after few seconds when no tuning was made, save current frequency to eeprom, this is to avoid its wear
	{
		if (freq != lastFreq) // this prevents overwriting with the same value if only turned on without tuning yet
		{
			eeprom_update_byte((unsigned char*)LAST_FREQ_ADDRESS, freq & 0xFF);
			eeprom_update_byte((unsigned char*)LAST_FREQ_ADDRESS + 1, freq >> 8);
			lastFreq = freq;
		}
	} else {
		lastFreqSaveCheck++;
	}

	switch (activeseg)
	{
		case 0:
			PORTD = ~segs[seg1];
			PORTB = 1;
		break;
		case 1:
			PORTD = ~segs[seg2];
			PORTB = 2;
			
			if (irLearningMode == 0 && isIrProgramSetMode == 0 && freqMhz < 10000)
			{
				PORTD &= ~_BV(_s_dot);
			}
		break;
		case 2:
			PORTD = ~segs[seg3];
			PORTB = 4;
			
			if (irLearningMode == 0 && isIrProgramSetMode == 0 && freqMhz >= 10000)
			{
				PORTD &= ~_BV(_s_dot);
			}
		break;
		case 3:
			PORTD = ~segs[seg4];
			PORTB = 8;
		break;
	}
}

ISR(COMPA_VECT)
{
	(void)irmp_ISR();
}
